/************************************************************************/
/* Effect for complete Deferred Shading                                 */
/************************************************************************/

/*
TODO:

	Complete Calculation in WorldSpace sucks...because of numerical problems...
	we MIGHT defer whole depth / DOF - Stuff into postprocessing (and bind one gbuffer to 
	post-fx)
	so we HAVE two coords. we could save the world-normal to (but wait... thats the same problem as before (we need the
	lightvec...and thats numerical stupid in either case...)

*/


/************************************************************************/
/* Variables                                                            */
/************************************************************************/

//Matrices
float4x4 matrxModelViewProj : ModelViewProjection;
float4x4 matrxWorld : World;

//DOF
//{ Near Blur Depth, Focal Plane Depth, Far Blur Depth, Bluriness Cutoff > Focal }
//are set dynamically
//float4 dofParams = { 1.291f, 2.0f, 4.30f, 1.0f };
float4 dofParams; // x = 1/(focusplanes - nearplanes), y = 1(farplane - focusplane),z = FocusPlan, w = clamp

//Samplers

sampler2D colorSampler = sampler_state {
	generateMipMap = false;
	MinFilter = LinearMipMapLinear;
	MagFilter = Linear;
	WrapS = Repeat;
	WrapT = Repeat;
	MaxAnisotropy = 16;
};

sampler2D normalSampler = sampler_state {
	GenerateMipMap = false;
	MinFilter = LinearMipMapLinear;
	MagFilter = Linear;
	WrapS = Repeat;
	WrapT = Repeat;
	MaxAnisotropy = 16;
};

samplerRECT bufferSampler0 = sampler_state {
	MinFilter = Nearest;
	MagFilter = Nearest;
	WrapS = ClampToEdge;
	WrapT = ClampToEdge;
};

samplerRECT bufferSampler1 = sampler_state {
	MinFilter = Nearest;
	MagFilter = Nearest;
	WrapS = ClampToEdge;
	WrapT = ClampToEdge;
};

samplerRECT bufferSampler2 = sampler_state {
	MinFilter = Nearest;
	MagFilter = Nearest;
	WrapS = ClampToEdge;
	WrapT = ClampToEdge;
};

//Shadow Sampler for VSDCT

sampler2D shadowMapVSDCT = sampler_state {
	CompareFunc = LEqual;
	CompareMode = CompareRToTexture;
	generateMipMap = false;
	MinFilter = Linear;
	MagFilter = Linear;
	WrapS = ClampToEdge;
	WrapT = ClampToEdge;
};

samplerCUBE indirectionSampler = sampler_state {
	GenerateMipMap = false;
	MinFilter = Linear;
	MagFilter = Linear;
	WrapS = ClampToEdge;
	WrapT = ClampToEdge;
};

//sampler3D lightResponseSampler = sampler_state {
//	GenerateMipMap = false;
//	MinFilter = Linear;
//	MagFilter = Linear;
//	WrapS = ClampToEdge;
//	WrapT = ClampToEdge;
//	WrapR = Repeat;
//};

sampler2D specularSampler = sampler_state {
	GenerateMipMap = false;
	MinFilter = LinearMipMapLinear;
	MagFilter = Linear;
	WrapS = Repeat;
	WrapT = Repeat;
	MaxAnisotropy = 16;
};

//Vars set by app for lightingpass
float specularity; //monochrome specularity
float shininess; //at least something that differs materials...
float4 sceneAmbient = { 0.2f, .2f, .2f, 1.0 }; //Ambient color set by scene
float4 lightPos = { .0f, 0.0f, .0f, 10.0f };
float4 lightColor = { 1.0f, 1.0f, 1.0f, 1.0f };
float lightIntensity;
float4 worldCamPos;

float2 polyOffset; 

//For drawing lightsource
static const float LSOURCE_EMISSIVE_COEFF = 39.78f;

//VSDCT - Parameter
float4   cLightQ;                       // [ Zf/(Zf-Zn), ZnZf/(Zf-Zn), 0, 0 ]

// application to vertex shader
struct app2vs
{
	float4 pos : POSITION; // position (object space)
	float3 normal : NORMAL; // normal (object space)
	float4 texcoord : TEXCOORD0; // texture coordinates
	float3 tangent : ATTR5; // tangent vector
	float3 bitangent : ATTR6; // bitangent vector
};

//Vertex Shader to Fragment Shader
struct vs2fs
{
	float4 hpos : POSITION; // position (clip space)
	float4 texcoord : TEXCOORD0; // texture coordinate + Depth!!! in texcoord.z
	float4 vpos : TEXCOORD1; // position (world-space)
	float3 normal : TEXCOORD2; // surface normal (world-space)
	float3 tangent : TEXCOORD3; // tangent vector (world-space)
	float3 bitangent : TEXCOORD4; // bitangent vector (world-space)
	//float3 view : TEXCOORD5; //View-Dir in TangenSpace
};

//Fragment Shader to MRTs
struct fs2mrt {
	float4 color0 : COLOR0; //Viewspace-POS
	float4 color1 : COLOR1; //Viewspace-Normal
	float4 color2 : COLOR2; //RGB-Albedo, Gloss
	//NOTE: we have depth available as depth-texture
};

struct simplestVsOutput
{
	float4 HPosition : POSITION;
	float2 UV : TEXCOORD0;
};

/************************************************************************/
/* GEOMETRY Stage                                                       */
/************************************************************************/

//vertex shader

vs2fs worldspaceVP(app2vs IN) {
	vs2fs OUT;

	//vertex position in object space
	float4 pos = float4(IN.pos.x,IN.pos.y,IN.pos.z,1.0);
	//vertex position in clip space
	float4 hpos = mul(matrxModelViewProj,pos);
	OUT.hpos = hpos;
	//copy texture coordinates
	OUT.texcoord.xy = IN.texcoord.xy;
	//compute world rotation only part
	float3x3 worldrot = float3x3(matrxWorld);

	//vertex position in view space (with model transformations)
	OUT.vpos = mul(matrxWorld, pos);
	
	//tangent space vectors in world space (with model trafos)
	float3 normal = mul(worldrot,IN.normal);
	OUT.normal.xyz = normal;

	//Storing depth in 4th comp of normal
	OUT.texcoord.z = hpos.z;

	float3 tangent = mul(worldrot,IN.tangent);
	OUT.tangent = tangent;

	float3 bitangent = mul(worldrot,IN.bitangent);
	OUT.bitangent = bitangent;

	return OUT;

}


simplestVsOutput simplestVP(float4 Position : POSITION, float2 TexCoord : TEXCOORD0) {
	simplestVsOutput OUT;
	OUT.HPosition = Position;
	//DEBUG
	//OUT.HPosition.z = 1.0f;
	OUT.UV = TexCoord;
	return OUT;
}

//fragment shader
fs2mrt fillGBufferFP(vs2fs IN) {

	fs2mrt OUT;
	//normal map
	float3 normal = f3tex2D(normalSampler, IN.texcoord.xy);

	//float3 normal = IN.normal.xyz;
	//color map
	float3 color = f3tex2D(colorSampler, IN.texcoord.xy);

	//specular/gloss map 
	float specular = tex2D(specularSampler, IN.texcoord.xy).x;
	
	//transform normal to world-space
	normal -= 0.5;
	//normal = normalize(normal.x*IN.tangent + normal.y*IN.bitangent + normal.z*IN.normal);

	normal = normalize(normal.x*IN.tangent + normal.y*IN.bitangent + normal.z*IN.normal);

#ifdef USEDOF
	//Calculate DOF - Params
	//NOTE: Using precalculated Values, see comments in common.fxh
	float depth = IN.texcoord.z;
	float f;
	if (depth < dofParams.z)
	{
		//Fragment is before focus plane
		// scale depth value range to [-1, 0]
		f = (depth - dofParams.z)*dofParams.x;
	} else {
		//Fragment is behind focus plane
		// scale depth range to [0, 1] 
		f = (depth - dofParams.z)*dofParams.y;
		// clamp the far blur to max. blurriness
		f = clamp(f, 0, dofParams.w);
	}
	//scale and bias into [0, 1] range
	f =  (f * 0.5f + 0.5f);
	OUT.color1.w = f;
#endif
	//note needed as we have floating point



	//NOTE: we could fill it in here w. DOF-Params
	//fill GBuffer
	//OUT.color0.xyz = IN.vpos.xyz;
	OUT.color0.xyz = IN.vpos.xyz;
	//OUT.color0.w = shininess;
	OUT.color1.xyz = normal;
	OUT.color2 = float4(color.xyz, specular * specularity);//gloss*specularity);

	return OUT;
}

//Depth Only Programs

struct vs2fs_depth
{
	float4 hpos : POSITION; // position (clip space)
};

vs2fs_depth depthOnlyVP(float4 position : POSITION) {
	vs2fs_depth OUT;
	OUT.hpos = mul(matrxModelViewProj,position);
	return OUT;
}

/************************************************************************/
/* Lighting Stage                                                       */
/************************************************************************/

float4 ambientOnlyFP(simplestVsOutput IN) : COLOR {
	float3 albedo = texRECT(bufferSampler2, IN.UV).xyz;
#ifndef USEHDR
	float dofDepth = texRECT(bufferSampler1, IN.UV).a;
#else
	static const float dofDepth = 1.0f;
#endif
	return float4(albedo * sceneAmbient, dofDepth);
}

//TODO: delete if not used anymore!!!
//Parameters for VSDCT NewFix
const float zbias_const = -0.00001;
//const float zbias_const = 0.001;
float2 xybias_const = { 4.8, 4.8 };

//Set by app, used by PCF
float2 inv_VSDCTSize;

//Shadow Switch
int doShadow = 0;

//NEw Poisson Shadow Filter
		static const float2 poisson[12] = {
			{-0.326212f, -0.405805f},
			{-0.840144f, -0.07358f},
			{-0.695914f, 0.457137f},
			{-0.203345f, 0.620716},
			{0.96234f, -0.194983f},
			{0.473434f, -0.480026f},
			{0.519456, 0.767022f},
			{0.185461f, -0.893124f},
			{0.507431f, 0.064425f},
			{0.89642f, 0.412458f},
			{-0.32194f, -0.932615f},
			{-0.791559f, -0.597705f}
		};
		float tapScale = 1.0f;
		//Eof Poisson Shadow Filter

float4 fullLightingFP(vs2fs_depth IN, float4 wPos : WPOS) : COLOR {

	//Retrieving data from GBuffers
	float4 pos = texRECT(bufferSampler0, wPos.xy);
	float3 normalVec;
	normalVec= texRECT(bufferSampler1, wPos.xy).xyz;

	//BUG: nicht normalisieren, wahrscheinlich ein NaN im FP- Buffer
	//normalVec = normalize(normalVec);
	float4 albedoGloss = texRECT(bufferSampler2, wPos.xy);

	//This one is now already in world-space
	float3 lightVec = lightPos.xyz - pos.xyz;
	float3 negLightVec = -1*lightVec;
	float3 viewVec = normalize(worldCamPos.xyz - pos.xyz);
	float att = saturate(1.0 - length(lightVec)/lightPos.w);
	lightVec = normalize(lightVec);
	float3 halfVec = normalize(viewVec + lightVec); 
	float ldn = max(0, dot(lightVec, normalVec));
	float hdn = pow(max(0, dot(halfVec, normalVec)), 30); //pos.w

	float3 iSpec = ldn*hdn*albedoGloss.w*lightColor.xyz * lightIntensity;

	float3 iDiff = ldn*albedoGloss.xyz*lightColor.xyz*lightIntensity;

#ifdef USESHADOWS
	half shadow;

	if (doShadow) {

	//Perform the VSDCT - Shadowing

		//See ShaderX3 - Article on Advanced Cube Texture Shadow Maps
		//float3 negLightVec = (-1*IN.LightVec).xyz;
		float3 LightVecAbs = abs(negLightVec);
		float4 shadowIndirectCoord;
		shadowIndirectCoord.xy = texCUBE(indirectionSampler, negLightVec).xw;

		//Adjust (like in MadMode-Mike Demo)
		//shadowIndirectCoord.xy -= float2(1.0f/1536.0f, 1.0f/1024.0f) * xybias_const * (frac(shadowIndirectCoord.xy * float2(3,2))*2 - 1);
		
		float MA = max(max(LightVecAbs.r, LightVecAbs.g), LightVecAbs.b);
		
		// project eye-space Z into windows - space Z
		//shadowIndirectCoord.b = ((-1.f/MA)*cLightQ.x + cLightQ.y+0.0001);
		const float c1 = 0.5*cLightQ.x;
		const float c2 = 0.5 + 0.5*cLightQ.y - zbias_const;

		shadowIndirectCoord.z = (1.0/MA)*c1 + c2;

#ifdef HQSFILTER
		half poissShadow = 0;

		for (int i=0; i< 12; i++) {
			float2 tapCoord = shadowIndirectCoord.xy + poisson[i] * tapScale * inv_VSDCTSize;// * (1.0f - shadowIndirectCoord.y);

			poissShadow += h1texcompare2D(shadowMapVSDCT, float3(tapCoord, shadowIndirectCoord.z))*(1.0f/12.0f);
		}
		shadow = poissShadow;
#else
		////Other, cheaper technique (as used in CG23 - release)
		//float2 offset = (float) (frac(wPos.xy * 0.5) > 0.25);
		//offset.y += offset.x;
		//if (offset.y > 1.1)
		//	offset.y = 0;

		//half shadow4tap = (

		//h1texcompare2D(shadowMapVSDCT, float3( shadowIndirectCoord.xy + (offset + float2(-1.5,0.5))*inv_VSDCTSize, shadowIndirectCoord.z)) +
		//h1texcompare2D(shadowMapVSDCT, float3( shadowIndirectCoord.xy + (offset + float2(0.5,0.5))*inv_VSDCTSize, shadowIndirectCoord.z)) +
		//h1texcompare2D(shadowMapVSDCT, float3( shadowIndirectCoord.xy + (offset + float2(-1.5,-1.5))*inv_VSDCTSize, shadowIndirectCoord.z)) +
		//h1texcompare2D(shadowMapVSDCT, float3( shadowIndirectCoord.xy + (offset + float2(0.5,-1.5))*inv_VSDCTSize, shadowIndirectCoord.z))

		//)*.25;
		//shadow = shadow4tap;


		half h = 0.25*(
                h1texcompare2D(shadowMapVSDCT, shadowIndirectCoord + float3(-0.3, 0.7,0)*1.5*float3(inv_VSDCTSize,0))  
              + h1texcompare2D(shadowMapVSDCT, shadowIndirectCoord + float3(-0.7,-0.3,0)*1.5*float3(inv_VSDCTSize,0))  
              + h1texcompare2D(shadowMapVSDCT, shadowIndirectCoord + float3( 0.3,-0.7,0)*1.5*float3(inv_VSDCTSize,0))  
              + h1texcompare2D(shadowMapVSDCT, shadowIndirectCoord + float3( 0.7, 0.3,0)*1.5*float3(inv_VSDCTSize,0)) 
             );
		shadow = h;
#endif

		
		
	//EOF SHADOWING
	} else {
		shadow = 1;
	}
#else
	static const float shadow = 1.0f;
#endif

	float4 result = { .0f, .0f, .0f, .0f };
	result.xyz = att*(iDiff*min(1, (shadow + 0.1)) + iSpec*shadow);
	//result.xyz = att*shadow*(iDiff + iSpec);
	return result;
}

float4 depthOnlyFP(vs2fs_depth IN) : COLOR {
	return float4(.0, .0, .0, .0);
}

float4 emissiveOnlyFP(vs2fs_depth IN) : COLOR {
	return float4(lightColor.xyz*lightIntensity*LSOURCE_EMISSIVE_COEFF, .0f);
}

float4 showBufferFP(simplestVsOutput IN) : COLOR {
	float4 result = { .0, .0, .0, 1.0 };
	result.xyz = texRECT(bufferSampler0, IN.UV).xyz;
	return result;
}

technique DeferredShading {

	pass CreateGBuffers {
		DepthTestEnable = true;
		AlphaTestEnable = false;
		StencilTestEnable = false;
		
		BlendEnable = false;
		ColorMask = bool4(true, true, true, true);
		DepthMask = false;
		DepthFunc = LEqual;
		CullfaceEnable = true;
		CullFace = Back;
		VertexProgram = compile vp40 worldspaceVP();
		FragmentProgram = compile fp40 fillGBufferFP();
	}

	pass depthOnly {

		DepthTestEnable = true;
		BlendEnable = false;
		DepthMask = true;
		DepthFunc = LEqual;
		ColorMask = bool4(false, false, false, false);
		CullFaceEnable = true;
		CullFace = Back;
		AlphaTestEnable = false;

		VertexProgram = compile vp40 depthOnlyVP();
		FragmentProgram = compile fp40 depthOnlyFP();
	}

	pass AmbientOnly {

		DepthFunc = Less;
		DepthTestEnable = false;
		DepthMask = false;
		BlendEnable = false;
		VertexProgram = compile vp40 simplestVP();
		FragmentProgram = compile fp40 ambientOnlyFP();

	}

	//We might unitize it with depthonly
	pass depthOnlyShadow {

		DepthTestEnable = true;
		BlendEnable = false;
		DepthMask = true;
		DepthFunc = LEqual;
		ColorMask = bool4(false, false, false, false);
		CullFaceEnable = true;
		CullFace = Front;
		AlphaTestEnable = false;

		PolygonOffsetFillEnable = true;
		PolygonOffset = polyOffset;

		VertexProgram = compile vp40 depthOnlyVP();
		FragmentProgram = compile fp40 depthOnlyFP();
	}

	pass fullLighting {
		
		//NOTE: Use Cullface=Back and DepthFunc=Less to get rid of occluders
		//		and use Cullface = front and DEpthFunc= Greater to get rid of "Behinder-Objects" not lit

		//-> if cam-pos is inside light volume, you HAVE to cull only front-faces

		CullfaceEnable = true;
		AlphaTestEnable = false;
		StencilTestEnable = false;
		BlendFunc = int2(One, One);	
		BlendEnable = true;
		DepthTestEnable = true;
		DepthMask = false;
		
		VertexProgram = compile vp40 depthOnlyVP();
		FragmentProgram = compile fp40 fullLightingFP();
	}

	pass drawLightSource {
		
		CullfaceEnable = true;
		CullFace = Back;
		AlphaTestEnable = false;
		StencilTestEnable = false;
		BlendFunc = int2(One, One);	
		BlendEnable = true;
		DepthTestEnable = true;
		DepthFunc = LEqual;
		DepthMask = false;
		
		VertexProgram = compile vp40 depthOnlyVP();
		FragmentProgram = compile fp30 emissiveOnlyFP();
	}

	pass showBuffer{
		DepthTestEnable = false;
		AlphaTestEnable = false;
		BlendEnable = false;
		DepthMask = true;
		VertexProgram = compile vp40 simplestVP();
		FragmentProgram = compile fp40 showBufferFP();
	}
}

//Used for debugging

//struct simplestVsOutput
//{
//	float4 HPosition : POSITION;
//	float2 UV : TEXCOORD0;
//};
//
//simplestVsOutput showBufferVP(float4 Position : POSITION, float2 TexCoord : TEXCOORD0) {
//	simplestVsOutput OUT;
//	OUT.HPosition = Position;
//	OUT.UV = TexCoord;
//	return OUT;
//}
//
//float4 showBufferFP(simplestVsOutput IN) : COLOR {
//	//float4 result = float4(.0, .0, .0, .0);
//	//float depth = tex2Dproj(bufferSampler2D, float4(IN.UV, 0, 1.0)).x;
//
//	//float sample = tex2D(bufferSampler2D, IN.UV.xy).r;
//
//
//	//result.x = sample;
//	float4 result = { .0, .0, .0, 1.0 };
//	result.xyz = texRECT(bufferSampler1, IN.UV).xyz;
//	//result.x = f1texRECT(bufferSampler1, IN.UV)*10;
//	//result.xyz = texRECT(bufferSampler0, float2(0.5, 0.5)).y;
//	//result /= 6;
//	//result.x = depth;
//	/*result = float4(.0, .0, .0, .0);
//	result.x = sample;*/
//	result.w = 1.0;
//	
//	
//
//	return result;
//}
//
//technique showBuffer {
//	pass {
//		DepthTestEnable = false;
//		AlphaTestEnable = false;
//		BlendEnable = false;
//		DepthMask = true;
//		VertexProgram = compile fp40 showBufferVP();
//		FragmentProgram = compile fp40 showBufferFP();
//	}
//}