//Techniques used in here are based on HDRLighting Example of DirectSDK

//DEBUG
float GAMMA = 1.0f;

//SAMPLERS

samplerRECT LightAccumSampler = sampler_state {
	MinFilter = Linear;
	MagFilter = Linear;
};

samplerRECT LuminanceSampler = sampler_state {
	//GenerateMipMap = false;
	MinFilter = Linear;
	MagFilter = Linear;
	/*WrapS = Repeat;
	WrapT = Repeat;*/
};

samplerRECT LuminanceAverage = sampler_state {
	//GenerateMipMap = false;
	MinFilter = Nearest;
	MagFilter = Nearest;
	/*WrapS = Repeat;
	WrapT = Repeat;*/
};

samplerRECT LuminanceAdapted = sampler_state {
	//GenerateMipMap = false;
	MinFilter = Linear;
	MagFilter = Linear;
	/*WrapS = Repeat;
	WrapT = Repeat;*/
};

samplerRECT BlurSource = sampler_state {
	//GenerateMipMap = false;
	MinFilter = Linear;
	MagFilter = Linear;
	/*WrapS = Repeat;
	WrapT = Repeat;*/
};

samplerRECT gbufferSampler = sampler_state {
	MinFilter = Nearest;
	MagFilter = Nearest;
	WrapS = ClampToEdge;
	WrapT = ClampToEdge;
};

static const float2 sampleOffset_4x4[16] = {
	float2(-1.5f, -1.5f),
	float2(-0.5f, -1.5f),
	float2(0.5f, -1.5f),
	float2(1.5f, -1.5f),

	float2(-1.5f, -0.5f),
	float2(-0.5f, -0.5f),
	float2(0.5f, -0.5f),
	float2(1.5f, -0.5f),

	float2(-1.5f, 0.5f),
	float2(-0.5f, 0.5f),
	float2(0.5f, 0.5f),
	float2(1.5f, 0.5f),

	float2(-1.5f, 1.5f),
	float2(-0.5f, 1.5f),
	float2(0.5f, 1.5f),
	float2(1.5f, 1.5f)
	
};

//Used for horizontally and vertically Gauss
//NOTE: weights are multiplied by 2
static const float2 offsetWeightsGauss15x15Sep[15] = {
	float2(-7.0f, 0.0174813),
	float2(-6.0f, 0.035994),
	float2(-5.0f, 0.0663181),
	float2(-4.0f, 0.10934),
	float2(-3.0f, 0.161314),
	float2(-2.0f, 0.212965),
	float2(-1.0f, 0.251589),
	float2(.0f, 0.265962),
	float2(1.0f, 0.251589),
	float2(2.0f, 0.212965),
	float2(3.0f, 0.161314),
	float2(4.0f, 0.10934),
	float2(5.0f, 0.0663181),
	float2(6.0f, 0.035994),
	float2(7.0f, 0.0174813)
};

//static const float  BRIGHT_PASS_THRESHOLD  = 5.0f;  // Threshold for BrightPass filter
static const float  BRIGHT_PASS_THRESHOLD  = 5.0f;  // Threshold for BrightPass filter
static const float  BRIGHT_PASS_OFFSET     = 20.0f; // Offset for BrightPass filter

// The per-color weighting to be used for luminance calculations in RGB order.
static const float3 LUMINANCE_VECTOR  = float3(0.2125f, 0.7154f, 0.0721f);

//static const float3 LUMINANCE_VECTOR  = float3(0.27f, 0.67f, 0.06f);

// The per-color weighting to be used for blue shift under low light.
static const float3 BLUE_SHIFT_VECTOR = float3(1.05f, 0.97f, 1.27f); 

//Tone Mapping Variables
float elapsedTime;
float MIDDLE_GRAY = 0.25;
static const float BLOOM_SCALE = 1.0f;



//INPUT _ OUTPUT STRUCS
struct downsampleVsOutput
{
	float4 HPosition : POSITION;
	float2 UV[4] : TEXCOORD0;
};

struct simplestVsOutput
{
	float4 HPosition : POSITION;
	float2 UV : TEXCOORD0;
};

struct simplestVsOutputPlusScaled
{
	float4 HPosition : POSITION;
	float2 UV : TEXCOORD0;
	float2 UV_scaled : TEXCOORD1;
};

/************************************************************************/
/* Vertex Programs                                                      */
/************************************************************************/

simplestVsOutput simplestVP(float4 Position : POSITION, float2 TexCoord : TEXCOORD0) {
	simplestVsOutput OUT;
	OUT.HPosition = Position;
	OUT.UV = TexCoord;
	return OUT;
}

simplestVsOutputPlusScaled simplestVP_scaled(float4 Position : POSITION, float2 TexCoord : TEXCOORD0) {
	simplestVsOutputPlusScaled OUT;
	OUT.HPosition = Position;
	OUT.UV = TexCoord;
	OUT.UV_scaled = TexCoord * 0.25;
	return OUT;
}

downsampleVsOutput DownSample4x4VP(float4 Position : POSITION, float2 TexCoord : TEXCOORD0) {

	downsampleVsOutput OUT;

	//Preparing Coords for downsampling by 4 with 4x4 box filter...
	float2 s = TexCoord;// + float2(-1.0, -1.0);
	OUT.HPosition = Position;

	OUT.UV[0] = s ;
	OUT.UV[1] = s + float2(2,0);
	OUT.UV[2] = s + float2(2,2);
	OUT.UV[3] = s + float2(0,2);


	//OUT.UV[0] = s + float2(-.5f, .5f);
	//OUT.UV[1] = s + float2(-.5f,-.5f);
	//OUT.UV[2] = s + float2(.5f,.5f);
	//OUT.UV[3] = s + float2(.5f, -.5f);

	return OUT;
}

/************************************************************************/
/* Fragment Programs                                                    */
/************************************************************************/


float4 DownSample4x4FP(downsampleVsOutput IN) : COLOR {
	
	//Downsampleing current LightAccumSampler
	//NOTE: still in in HDR-Format
	half3 sum = half3(0,0,0);
	sum += h3texRECT(LightAccumSampler, IN.UV[0]);
	sum += h3texRECT(LightAccumSampler, IN.UV[1]);
	sum += h3texRECT(LightAccumSampler, IN.UV[2]);
	sum += h3texRECT(LightAccumSampler, IN.UV[3]);
	sum *= (half) 0.25f;

	return float4(sum, 1.0f);
}


float4 ConvertToLuminance_FP(simplestVsOutput IN) : COLOR
{
    half3 vSample = h3texRECT(LightAccumSampler, IN.UV);

	half vSampleLum;
    //vSampleLum = log(dot(vSample, LUMINANCE_VECTOR)+0.0001f);
	vSampleLum = (dot(vSample, LUMINANCE_VECTOR)+0.0001f);

	//R-Channel only
    return float4(vSampleLum, .0f, .0f, .0f);
}

float4 SampleLumIterativeFP(simplestVsOutput IN) : COLOR
{
    half fResampleSum = 0.0f; 
    
    for(int iSample = 0; iSample < 16; iSample++)
    {
        // Compute the sum of luminance throughout the sample points
        fResampleSum += h1texRECT(LuminanceSampler, IN.UV + sampleOffset_4x4[iSample]);
    }
    
    // Divide the sum to complete the average
    fResampleSum *= 0.0625;

    return float4(fResampleSum, .0, .0, .0);
}

float4 SampleLumFinalFP(simplestVsOutput IN) : COLOR
{
    half fResampleSum = 0.0f;
    
    for(int iSample = 0; iSample < 16; iSample++)
    {
        // Compute the sum of luminance throughout the sample points
        fResampleSum += h1texRECT(LuminanceSampler, IN.UV+sampleOffset_4x4[iSample]);
    }
    
    // Divide the sum to complete the average, and perform an exp() to complete
    // the average luminance calculation
    //fResampleSum = exp(fResampleSum/16);
	fResampleSum = (fResampleSum/16);
    
    return float4(fResampleSum, .0, .0, .0);
}

float4 CalculateAdaptedLumFP(simplestVsOutput IN) : COLOR
{
    half fAdaptedLum = h1texRECT(LuminanceAdapted, float2(0.5f, 0.5f));
    half fCurrentLum = h1texRECT(LuminanceAverage, float2(0.5f, 0.5f));
    
    // The user's adapted luminance level is simulated by closing the gap between
    // adapted luminance and current luminance by 2% every frame, based on a
    // 30 fps rate. This is not an accurate model of human adaptation, which can
    // take longer than half an hour.
    half fNewAdaptation = fAdaptedLum + (fCurrentLum - fAdaptedLum) * ( 1 - pow( 0.97f, 30 * elapsedTime ) );
    return float4(fNewAdaptation, .0, .0, .0);
	//return float4(fCurrentLum, fCurrentLum, fCurrentLum, 1.0f);
}

float4 BrightPassFilterFP(simplestVsOutput IN) : COLOR 
{
	half3 vSample = h3texRECT( LightAccumSampler, IN.UV );
	half  fAdaptedLum = h1texRECT( LuminanceAdapted, float2(0.5f, 0.5f) );
	
	// Determine what the pixel's value will be after tone-mapping occurs
	vSample.rgb *= (half) MIDDLE_GRAY/(fAdaptedLum + 0.001f);
	
	//// Subtract out dark pixels
	vSample.rgb -= (half) BRIGHT_PASS_THRESHOLD;
	
	// Clamp to 0
	vSample = max(vSample, 0.0f);

	vSample.rgb /= ((half) BRIGHT_PASS_OFFSET+vSample);

    
	return float4(vSample, 1.0f);
}

float4 BlurGauss_Horiz_FP(simplestVsOutput IN) : COLOR
{
    
    half3 vSample = 0.0f;
    half3 vColor = 0.0f;
        
	float2 vSamplePosition;
    
    // Perform a one-directional gaussian blur
    for(int iSample = 0; iSample < 15; iSample++)
    {
        vSamplePosition = IN.UV + float2(offsetWeightsGauss15x15Sep[iSample].x, .0f);
        vColor = h3texRECT(BlurSource, vSamplePosition);
        vSample += (half) offsetWeightsGauss15x15Sep[iSample].y*vColor;
    }
    
    return float4( vSample, 1.0f);
}

float4 BlurGauss_Vert_FP(simplestVsOutput IN) : COLOR
{
    
    half3 vSample = 0.0f;
    half3 vColor = 0.0f;
        
	float2 vSamplePosition;
    
    // Perform a one-directional gaussian blur
    for(int iSample = 0; iSample < 15; iSample++)
    {
        vSamplePosition = IN.UV + float2(.0f, offsetWeightsGauss15x15Sep[iSample].x);
        vColor = h3texRECT(BlurSource, vSamplePosition);
        vSample += (half) offsetWeightsGauss15x15Sep[iSample].y*vColor; //DEBUG
    }
    
    return float4(vSample, 1.0f);
}

float4 FinalSceneFP(simplestVsOutputPlusScaled IN) : COLOR
{
    half3 vSample = h3texRECT(LightAccumSampler, IN.UV);
	//TODO: calculate scaled UV-Coords in own VertexShader and fix it
    half3 vBloom = h3texRECT(BlurSource, IN.UV_scaled);
	half fAdaptedLum = h1texRECT(LuminanceAdapted, float2(0.5f, 0.5f));
    
	//BLUESHIFT
	// Define a linear blending from -1.5 to 2.6 (log scale) which
	// determines the lerp amount for blue shift
    half fBlueShiftCoefficient = 1.0f - (fAdaptedLum + 1.5)/4.1;
    fBlueShiftCoefficient = saturate(fBlueShiftCoefficient);

	// Lerp between current color and blue, desaturated copy
    half3 vRodColor = dot( vSample, (half3) LUMINANCE_VECTOR ) * (half3) BLUE_SHIFT_VECTOR;
    vSample.rgb = lerp( vSample, vRodColor, fBlueShiftCoefficient );
    
	half3 vRodColorBloom = dot( vBloom, (half3) LUMINANCE_VECTOR ) * (half3) BLUE_SHIFT_VECTOR;
	vBloom.rgb = lerp( vBloom, vRodColorBloom, fBlueShiftCoefficient );

	//EOF BLUESHIFT

    // Map the high range of color values into a range appropriate for
    // display, taking into account the user's adaptation level, and selected
    // values for for middle gray and white cutoff.
	vSample.rgb *= (half) MIDDLE_GRAY/(fAdaptedLum + 0.001f);
	vSample.rgb /= (1.0f+vSample);

    // Add the star and bloom post processing effects
    vSample += (half) BLOOM_SCALE * vBloom;
    
	//Finally fetch the DOF-Alpha value, and save it to destination alpha
	//Might still have to do this in here...
#ifdef USEDOF
	half alphaDof = h4texRECT(gbufferSampler, IN.UV).a;
#else
	static const float alphaDof = 1.0f;
#endif
	//Now done later
	//DEBUG GAMMA CORR
	//vSample = pow(vSample, GAMMA);
	//EOF DEBUG
	//vSample = vBloom;
    return float4( vSample, alphaDof);
	//return vBloom;
}

/************************************************************************/
/* Techniques                                                           */
/************************************************************************/

technique ToneMapping {
	
	pass DownSample4x4 {
		DepthTestEnable = false;
		BlendEnable = false;
		
		VertexProgram = compile vp40 DownSample4x4VP();
		FragmentProgram = compile fp40 DownSample4x4FP();
	}

	pass ConvertToLuminance {
		DepthTestEnable = false;
		BlendEnable = false;
		
		VertexProgram = compile vp40 simplestVP();
		FragmentProgram = compile fp40 ConvertToLuminance_FP();
	}

	//NOTE: simple Downscale does not work, it gets too dark, 
	//		and contrast seems to be unreasonable
	pass IterativeLuminance {
		DepthTestEnable = false;
		BlendEnable = false;
		VertexProgram = compile vp40 simplestVP();
		FragmentProgram = compile fp40 SampleLumIterativeFP();
	}
		
	pass FinalLuminance {
		DepthTestEnable = false;
		BlendEnable = false;
		
		VertexProgram = compile vp40 simplestVP();
		FragmentProgram = compile fp40 SampleLumFinalFP();
	}
	
	pass LuminanceAdaptation {
		DepthTestEnable = false;
		BlendEnable = false;
		
		VertexProgram = compile vp40 simplestVP();
		FragmentProgram = compile fp40 CalculateAdaptedLumFP();
	}
	
	pass BrightPassFilter {
		DepthTestEnable = false;
		BlendEnable = false;
		
		VertexProgram = compile vp40 simplestVP();
		FragmentProgram = compile fp40 BrightPassFilterFP();
	}

	pass BlurGaussSepHoriz {
		DepthTestEnable = false;
		BlendEnable = false;
		
		VertexProgram = compile vp40 simplestVP();
		FragmentProgram = compile fp40 BlurGauss_Horiz_FP();
	}

	pass BlurGaussSepVert {
		DepthTestEnable = false;
		BlendEnable = false;
		
		VertexProgram = compile vp40 simplestVP();
		FragmentProgram = compile fp40 BlurGauss_Vert_FP();
	}

	pass FinalScene {
		DepthTestEnable = false;
		BlendEnable = false;
		
		VertexProgram = compile vp40 simplestVP_scaled();
		FragmentProgram = compile fp40 FinalSceneFP();
	}
}